@@ -1,4 +1,4 @@
-# Copyright (c) 2019 - 2024, Adrian Dusa
+# Copyright (c) 2019 - 2025, Adrian Dusa
 # All rights reserved.
 # 
 # Redistribution and use in source and binary forms, with or without
@@ -24,22 +24,41 @@
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-`recode` <- function(x, rules, cut, values = NULL, ...) {
+`recode` <- function(x, rules = NULL, cut = NULL, values = NULL, ...) {
     UseMethod("recode")
 }
-`recode.declared` <- function(x, rules, cut, values = NULL, ..., na_values = TRUE) {
+`recode.declared` <- function(x, rules = NULL, cut = NULL, values = NULL, ...) {
+    dots <- list(...)
     na_index <- attr(x, "na_index")
+    na_values <- attr(x, "na_values")
+    na_range <- attr(x, "na_range")
+    xlabels <- attr(x, "labels", exact = TRUE)
     attributes(x) <- NULL
-    if (!is.null(na_index) & isTRUE(na_values)) {
-        nms <- names(na_index)
-        if (possibleNumeric(nms)) {
-            nms <- asNumeric(nms)
+    labels <- dots$labels
+    x <- recode(x = x, rules = rules, cut = cut, values = values)
+    if (is.null(names(labels))) {
+        values <- sort(unique(x))
+        if (length(values) == length(labels)) {
+            names(values) <- labels
+            labels <- values
         }
-        x[na_index] <- nms
     }
-    NextMethod()
+    attr(x, "na_index") <- na_index
+    attr(x, "na_values") <- na_values
+    attr(x, "na_range") <- na_range
+    if (!is.null(xlabels)) {
+        if (!is.null(na_values)) {
+            xlabels <- xlabels[is.element(labels, na_values)]
+        }
+        else if (!is.null(na_range)) {
+            xlabels <- xlabels[xlabels >= na_range[1] & xlabels <= na_range[2]]
+        }
+    }
+    attr(x, "labels") <- c(labels, xlabels)
+    class(x) <- c("declared", class(x))
+    return(x)
 }
-`recode.default` <- function(x, rules, cut, values = NULL, ...) {
+`recode.default` <- function(x, rules = NULL, cut = NULL, values = NULL, ...) {
     if (missing(x)) {
         stopError("Argument 'x' is missing.")
     }
@@ -55,7 +74,6 @@
     factor.levels     <- splitstr(dots$levels)
     factor.labels     <- splitstr(dots$labels)
     factor.ordered    <- FALSE
-    declared <- inherits(x, "declared")
     if (is.element("ordered", names(dots))) {
         factor.ordered <- dots$ordered
     }
@@ -93,7 +111,10 @@
         if (length(c(seqfrom, seqto)) < 2) return(NULL)
         return(seq(seqfrom, seqto))
     }
-    if (missing(cut)) {
+    if (is.null(cut)) {
+        if (is.null(rules)) {
+            stopError("At least one argument 'rules' or 'cut' should be provided.")
+        }
         rules <- gsub(
             "\n|\t", "", gsub(
                 "'", "", gsub(
@@ -165,7 +186,7 @@
                 torecode <- getFromRange(from[i], to[i], uniques, xisnumeric)
                 if (!is.null(torecode)) {
                     vals <- uniques[torecode]
-                    temp[x %in% vals] <- newval[i]
+                    temp[is.element(x, vals)] <- newval[i]
                     recoded <- c(recoded, vals)
                 }
             }
@@ -252,24 +273,17 @@
             }
         }
         else {
-            sx <- sort(x)
-            minx <- sx[1]
-            maxx <- sx[length(x)]
             if (is.character(x) & is.numeric(cut)) {
                 insidex <- FALSE
             }
-            else {
-                insidex <- logical(length(cut))
-                for (i in seq(length(cut))) {
-                    insidex[i] <- cut[i] >= minx & cut[i] <= maxx
-                }
+            else if (is.character(x) & is.character(cut)) {
+                insidex <- is.element(cut, x[!is.na(x)])
+            } else {
+                insidex <- cut >= min(x, na.rm = TRUE) & cut <= max(x, na.rm = TRUE)
             }
         }
         if (!all(insidex)) {
             message <- "Cut value(s) outside the input vector."
-            if (declared) {
-                message <- paste(message, "Consider using undeclare() before recoding.")
-            }
             stopError(message)
         }
         if (is.factor(x)) {
@@ -282,10 +296,14 @@
             }
         }
         else {
+            nax <- which(is.na(x))
             temp <- rep(values[1], length(x))
             for (i in seq(length(cut))) {
                 temp[x > cut[i]] = values[i + 1]
             }
+            if (length(nax) > 0) {
+                temp[nax] <- NA
+            }
         }
         if (!is.null(factor.labels) && length(factor.labels) == 0 && is.numeric(cut)) {
             factor.labels <- values
